import math
from servicepytan.utils import request_json, get_timezone_by_file, endpoint_url, request_json_with_retry

def get_report_categories(conn=None):
    """Get a list of report categories"""
    return request_json(endpoint_url('reporting', 'report-categories', conn=conn), conn=conn)

def get_report_list(report_category, conn=None):
    """Get a list of reports for a given report category"""
    return request_json(endpoint_url('reporting', f'report-category/{report_category}/reports', conn=conn), conn=conn)

def get_dynamic_set_list(dynamic_set_id,conn=None):
    """Get a list of dynamic sets"""
    return request_json(endpoint_url('reporting', f'dynamic-value-sets/{dynamic_set_id}', conn=conn), conn=conn)

class Report:
  """Primary class for retrieving Reporting Endpoint Data.

  Attributes:
      category: A string representing the report category. Find list of categories with get_report_categories().
      report_id: A string representing the report id. Find list of report_id using get_report_list().
      conn: a dictionary containing the credential config.
  """
  def __init__(self, category, report_id, conn=None):
    """Inits DataService with configuration file and authentication settings."""
    self.conn = conn
    # self.timezone = get_timezone_by_file(conn)
    self.category = category
    self.report_id = report_id
    self.params = {"parameters": []}
    self.metadata = self.get_metadata()

  def add_params(self, name, value):
    """add a parameter to the report"""
    param_keys = [param["name"] for param in self.params["parameters"]]
    if name in param_keys:
      print(f"Parameter '{name}' already exists. Updating value from '{self.params['parameters'][param_keys.index(name)]['value']}' to '{value}'...")
      self.update_params(name, value)
    else:
      self.params["parameters"].append({"name": name, "value": value})

  def update_params(self, name, value):
    """update a parameter in the report"""
    param_keys = [param["name"] for param in self.params["parameters"]]
    if name in param_keys:
      self.params["parameters"][param_keys.index(name)]["value"] = value
    else:
      print(f"Parameter '{name}' does not exist. Adding parameter...")
      self.add_params(name, value)

  def get_params(self):
    """get report parameters"""
    return self.params

  def get_metadata(self):
    """get report metadata"""
    endpoint = f"report-category/{self.category}/reports/{self.report_id}"
    url = endpoint_url("reporting",endpoint, conn=self.conn)
    return request_json_with_retry(url, conn=self.conn)

  def show_param_types(self):
    """show parameter types"""
    for param in self.metadata["parameters"]:
      dynamic_set_id = ""
      required = "[ ]"
      if param["isRequired"]:
        required = "[*]"
      if param['acceptValues']:
        dynamic_set_id = f" (dynamicSetId: {param['acceptValues']['dynamicSetId']})"
      print(f"{required} - {param['name']}: {param['dataType']}, {dynamic_set_id}")

  def get_data(self, params="", page=1, page_size=5000):
    """get report data"""
    if params == "":
      params = self.params
    options = {"page": page, "pageSize": page_size, "includeTotal": True}
    endpoint = f"report-category/{self.category}/reports/{self.report_id}/data"
    url = endpoint_url("reporting",endpoint, conn=self.conn)
    return request_json_with_retry(url, options=options, json_payload=params, 
              conn=self.conn, request_type="POST")
  
  def get_all_data(self, params="", page_size=5000):
    """get all report data"""
    page = 1
    data = []
    fields = []
    if params == "":
      params = self.params
    print("Getting first page of data...")
    response = self.get_data(params, page=page, page_size=page_size)
    data.extend(response["data"])
    fields.extend(response["fields"])
    total = response["totalCount"]
    has_more = response["hasMore"]
    print(f"Retrieved {len(data)} of {total} records...")
    requests_needed = math.ceil(total / page_size)
    mins_to_complete = requests_needed * 5
    updated_page_size = page_size
    if mins_to_complete > 60 or requests_needed > 2:
      init_page_size = updated_page_size
      if page_size < 5000 and math.ceil(total / 5000) < 12:
        print("Setting page size to 5000 to speed up report retrieval...")
        updated_page_size = 5000
        requests_needed = 1 + math.ceil((total - init_page_size) / updated_page_size)
      else:
        print(f"This request will take at least {mins_to_complete/60} hours to complete.")
        print("Limit the parameters to reduce the number of requests and try again.")
        return {"error": "Too many requests. Try again with fewer parameters."}
    while has_more:
      page += 1
      print(f"Getting page {page} of {requests_needed}...")
      response = self.get_data(params, page=page, page_size=updated_page_size)
      if(len(response["data"]) == 0):
        print("No more data to retrieve.")
        break
      data.extend(response["data"])
      print(f"Retrieved {len(data)} sof {total} records...")
      has_more = response["hasMore"]
    return {"data": data, "fields": fields}


"""Dates Module: Converting Dates to UTC format"""

from dateutil.parser import parse
import pytz

def _convert_date_to_api_format(date, timezone=""):
  """Converts a date into ISO format compatible with API endpoint parameters.

  The function will either format the date as is if no timezone is provided or convert to UTC from
  the timezone parameter. The function accepts BOTH a string that it will try to parse into a 
  datetime object. If a datetime object is also acceptable.

  Args:
      date: A string or datetime object
      timezone: A string using a Timezone DB abbreviation (e.g. 'America/New_York' = Eastern Time)

  Returns:
      A date string in ISO format
  """  
  # Checks the date string for format and parses if not date object
  if isinstance(date, str):
    parsed_date = _parse_date_string(date)
  else:
    parsed_date = date
  
  # If no timezone is provided then assume it is provided in UTC
  # Otherwise, perform the conversion from the stated TZ to UTC
  if timezone != "":
    parsed_date = _change_timezones(parsed_date, timezone)

  formatted_date = _format_date_to_iso_format(parsed_date)
  return formatted_date

def _parse_date_string(date_string):
  """Parse date string to datetime object"""
  return parse(date_string)

def _change_timezones(datetime_object, timezone):
    """Convert a datetime object from one timezone to UTC"""
    timezone = pytz.timezone(timezone)
    parsed_date = timezone.localize(datetime_object)
    parsed_date = _convert_datetime_to_utc(parsed_date)
    return parsed_date

def _format_date_to_iso_format(datetime_object):
  """Format datetime object to ISO format"""
  return datetime_object.strftime('%Y-%m-%dT%H:%M:%SZ')

def _convert_datetime_to_utc(datetime_object):
  """Convert datetime object to UTC"""
  return datetime_object.astimezone(pytz.UTC)


"""Authenticating with ServiceTitan API"""

import requests
import json
import os
from dotenv import load_dotenv

from servicepytan import URL_ROOT, AUTH_ROOT

AUTH_VARIABLES = [
    'SERVICETITAN_APP_KEY',
    'SERVICETITAN_TENANT_ID',
    'SERVICETITAN_CLIENT_ID',
    'SERVICETITAN_CLIENT_SECRET',
    'SERVICETITAN_APP_ID',
    'SERVICETITAN_TIMEZONE'
]

def servicepytan_connect(
    app_key:str=None, tenant_id:str=None, client_id:str=None, 
    client_secret:str=None, app_id:str=None, timezone:str="UTC", config_file:str=None):
    
    auth_config_object = {
            "SERVICETITAN_APP_KEY": app_key,
            "SERVICETITAN_TENANT_ID": tenant_id,
            "SERVICETITAN_CLIENT_ID": client_id,
            "SERVICETITAN_CLIENT_SECRET": client_secret,
            "SERVICETITAN_APP_ID": app_id,
            "SERVICETITAN_TIMEZONE": timezone
    }


    # First check if the config_file is provided
    if config_file:
        print("Setting auth config from file...")
        f = open(config_file)
        creds = json.load(f)
        for var in AUTH_VARIABLES:
            auth_config_object[var] = creds.get(var, '')
        f.close()

    # If not, check if the environment variables are set
    elif not app_key or not tenant_id or not client_id or not client_secret or not app_id:
        load_dotenv()
        print("Auth config not provided, loading from environment variables...")
        for var in AUTH_VARIABLES:
            auth_var = os.environ.get(var)
            if auth_var:
                auth_config_object[var] = auth_var
            else:
                print(f"Environment variable {var} not found or provided in function. Defualting to empty string.")
                auth_config_object[var] = ''

    return auth_config_object

def request_auth_token(client_id, client_secret):
  """Fetches Auth Token.

  Retrieves authentication token for completing a request against the API

  Args:
      client_id: String, provided from the integration settings
      client_secret: String, provided from the integration settings

  Returns:
      Authentication token

  Raises:
      TBD
  """

  url = f"{AUTH_ROOT}/connect/token"

  querystring = {"Content-Type":"application/x-www-form-urlencoded"}

  payload = f"grant_type=client_credentials&client_id={client_id}&client_secret={client_secret}"
  headers = {"Content-Type": "application/x-www-form-urlencoded"}

  response = requests.request("POST", url, data=payload, headers=headers, params=querystring)

  return json.loads(response.text)

def get_auth_token(conn):
  """Fetches Auth Token using the config_file.

  Retrives the CLIENT_ID and CLIENT_SECRET entries from config_file.

  Args:
      config_file: String, path to the config file defaults to 'servicepytan_config.json'

  Returns:
      Authentication token

  Raises:
      TBD
  """
  # Read File
  client_id = conn['SERVICETITAN_CLIENT_ID']
  client_secret = conn['SERVICETITAN_CLIENT_SECRET']
  return request_auth_token(client_id, client_secret)["access_token"]

def get_app_key(conn):
  """Fetches App Key from the config_file.

  Retrives the APP_KEY entry from config_file.

  Args:
      config_file: String, path to the config file defaults to 'servicepytan_config.json'

  Returns:
      App Key

  Raises:
      TBD
  """
  app_key = conn['SERVICETITAN_APP_KEY']
  return app_key

def get_tenant_id(conn):
  """Fetches Tenant ID from the config_file.

  Retrives the TENANT_ID entry from config_file.

  Args:
      config_file: String, path to the config file defaults to 'servicepytan_config.json'

  Returns:
      Tenant ID

  Raises:
      TBD
  """
  tenant_id = conn['SERVICETITAN_TENANT_ID']
  return tenant_id 

def get_auth_headers(conn):
  """Generates the Authentication Headers for each API request

  Creates an object that includes the auth token and app key formatted to create the auth headers.

  Args:
      config_file: String, path to the config file defaults to 'servicepytan_config.json'

  Returns:
      Object

  Raises:
      TBD
  """
  return {
      "Authorization": get_auth_token(conn),
      "ST-App-Key": get_app_key(conn)
  }